/*
 * lcdtest: LCD monitor test pattern generator
 *
 * Main program
 * $Id: lcdtest.c 18 2005-04-23 19:10:41Z eric $
 * Copyright 2005 Eric Smith <eric@brouhaha.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.  Note that permission is
 * not granted to redistribute this program under the terms of any
 * other version of the General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 */


#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <SDL/SDL.h>


char *progname;


void help (FILE *f)
{
  fprintf (f, "\n");
  fprintf (f, "commands:\n");
  fprintf (f, "  Change foreground color:\n");
  fprintf (f, "      r:  red\n");
  fprintf (f, "      g:  green\n");
  fprintf (f, "      b:  blue\n");
  fprintf (f, "      w:  white\n");
  fprintf (f, "      k:  black\n");
  fprintf (f, "  Change pattern:\n");
  fprintf (f, "      s:  solid fill\n");
  fprintf (f, "      h:  horizontal lines\n");
  fprintf (f, "      v:  vertical lines\n");
  fprintf (f, "      c:  crosshatch\n");
  fprintf (f, "      a:  white vertical lines with two pixel pitch\n");
  fprintf (f, "  arrow keys: move lines in arrow direction\n");
  fprintf (f, "  plus:   increase spacing between lines\n");
  fprintf (f, "  minus:  decrease spacing between lines\n");
  fprintf (f, "  0..9:   select predefined line spacings\n");
  fprintf (f, "  q or Escape: exit program\n");
}


void usage (FILE *f)
{
  fprintf (f, "lcdtest -- Copyright 2005 Eric Smith <eric@brouhaha.com>\n");
  fprintf (f, "http://www.brouhaha.com/~eric/software/lcdtest/");
  fprintf (f, "\n");
  fprintf (f, "usage:  %s [options]\n", progname);
  fprintf (f, "options:\n");
  fprintf (f, "    --help         get list of command keys\n");
  fprintf (f, "    --window       display in a window rather than full screen\n");
  fprintf (f, "    -x <count>     use a video mode with the specified horizontal resolution\n");
  fprintf (f, "    -y <count>     use a video mode with the specified vertical resolution\n");
  fprintf (f, "    --list-modes   list available video modes (including resolution)\n");
}


void fatal (int ret, char *format, ...)
{
  va_list ap;

  fprintf (stderr, "fatal error: ");
  va_start (ap, format);
  vfprintf (stderr, format, ap);
  va_end (ap);
  if (ret == 1)
    usage (stderr);
  exit (ret);
}


typedef struct
{
  uint8_t red;
  uint8_t green;
  uint8_t blue;
} my_color_t;

my_color_t test_colors [] =
{
  { 0xff, 0x00, 0x00 },
  { 0x00, 0xff, 0x00 },
  { 0x00, 0x00, 0xff },
  { 0xff, 0xff, 0xff },
  { 0x80, 0x80, 0x80 },
  { 0x00, 0x00, 0x00 }
};


SDLKey wait_key (void)
{
  SDL_Event event;

  while (1)
    {
      SDL_WaitEvent (& event);
      switch (event.type)
	{
	case SDL_KEYDOWN:
	  return event.key.keysym.sym;
	case SDL_QUIT:
	  exit (0);
	}
    }
}


void lock_screen (SDL_Surface *screen)
{
  if (SDL_MUSTLOCK (screen))
    if (SDL_LockSurface (screen) < 0)
      fatal (2, "can't lock screen\n");
}


void unlock_screen (SDL_Surface *screen)
{
  if (SDL_MUSTLOCK (screen))
    SDL_UnlockSurface (screen);
}


void fill_screen (SDL_Surface *screen, uint32_t color)
{
  SDL_Rect rect;
  rect.x = 0;
  rect.y = 0;
  rect.w = screen->w;
  rect.h = screen->h;
  if (SDL_FillRect (screen, & rect, color) < 0)
    fatal (2, "SDL_FillRect failed; %s\n", SDL_GetError ());
}


void draw_horiz_lines (SDL_Surface *screen,
		       uint32_t color,
		       int pitch,
		       int offset)
{
  SDL_Rect rect;
  rect.x = 0;
  rect.w = screen->w;
  rect.h = 1;
      
  for (rect.y = offset; rect.y < screen->h; rect.y += pitch)
    if (SDL_FillRect (screen, & rect, color) < 0)
      fatal (2, "SDL_FillRect failed; %s\n", SDL_GetError ());
}


void draw_vert_lines (SDL_Surface *screen,
		      uint32_t color,
		      int pitch,
		      int offset)
{
  SDL_Rect rect;
  rect.y = 0;
  rect.h = screen->h;
  rect.w = 1;
      
  for (rect.x = offset; rect.x < screen->w; rect.x += pitch)
    if (SDL_FillRect (screen, & rect, color) < 0)
      fatal (2, "SDL_FillRect failed; %s\n", SDL_GetError ());
}


uint32_t black;


void pattern_solid (SDL_Surface *screen, uint32_t color)
{
  lock_screen (screen);
  fill_screen (screen, color);
  unlock_screen (screen);
  SDL_UpdateRect (screen, 0, 0, screen->w, screen->h);
}

void pattern_horiz_lines (SDL_Surface *screen,
			  uint32_t color,
			  int v_pitch,
			  int v_offset)
{
  lock_screen (screen);
  fill_screen (screen, black);
  draw_horiz_lines (screen, color, v_pitch, v_offset);
  unlock_screen (screen);
  SDL_UpdateRect (screen, 0, 0, screen->w, screen->h);
}

void pattern_vert_lines (SDL_Surface *screen,
			 uint32_t color,
			 int h_pitch,
			 int h_offset)
{
  lock_screen (screen);
  fill_screen (screen, black);
  draw_vert_lines (screen, color, h_pitch, h_offset);
  unlock_screen (screen);
  SDL_UpdateRect (screen, 0, 0, screen->w, screen->h);
}

void pattern_crosshatch (SDL_Surface *screen,
			 uint32_t color,
			 int h_pitch,
			 int h_offset,
			 int v_pitch,
			 int v_offset)
{
  lock_screen (screen);
  fill_screen (screen, black);
  draw_horiz_lines (screen, color, v_pitch, v_offset);
  draw_vert_lines (screen, color, h_pitch, h_offset);
  unlock_screen (screen);
  SDL_UpdateRect (screen, 0, 0, screen->w, screen->h);
}


typedef enum { solid, vert_lines, horiz_lines, crosshatch } pattern_t;


#define set_pitch(x) do \
                       if ((pattern != solid) && ((x) < screen->w) && ((x) < screen->h)) \
                         pitch = (x); \
                     while (0)


void main_loop (SDL_Surface *screen)
{
  pattern_t pattern = solid;
  uint32_t fg_color;
  int pitch = 32;
  int h_offset = 0;
  int v_offset = 0;
  SDLKey key;

  black = SDL_MapRGB (screen->format, 0x00, 0x00, 0x00);

  fg_color = SDL_MapRGB (screen->format, 0xff, 0x00, 0x00);

  SDL_EnableKeyRepeat (SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

  while (1)
    {
      if (h_offset >= pitch)
	h_offset %= pitch;
      if (v_offset >= pitch)
	v_offset %= pitch;

      switch (pattern)
	{
	case solid:
	  pattern_solid (screen, fg_color);
	  break;
	case vert_lines:
	  pattern_vert_lines (screen, fg_color, pitch, h_offset);
	  break;
	case horiz_lines:
	  pattern_horiz_lines (screen, fg_color, pitch, v_offset);
	  break;
	case crosshatch:
	  pattern_crosshatch (screen, fg_color, pitch, h_offset, pitch, v_offset);
	  break;
	}

      key = wait_key ();
      switch (key)
	{
	case SDLK_ESCAPE:
	case SDLK_q:
	  exit (0);
	case SDLK_r:
	  fg_color = SDL_MapRGB (screen->format, 0xff, 0x00, 0x00);
	  break;
	case SDLK_g:
	  fg_color = SDL_MapRGB (screen->format, 0x00, 0xff, 0x00);
	  break;
	case SDLK_b:
	  fg_color = SDL_MapRGB (screen->format, 0x00, 0x00, 0xff);
	  break;
	case SDLK_w:
	  fg_color = SDL_MapRGB (screen->format, 0xff, 0xff, 0xff);
	  break;
	case SDLK_k:
	  fg_color = SDL_MapRGB (screen->format, 0x00, 0x00, 0x00);
	  break;
	case SDLK_a:
	  fg_color = SDL_MapRGB (screen->format, 0xff, 0xff, 0xff);
	  pattern = vert_lines;
	  pitch = 2;
	  h_offset = 0;
	  break;
	case SDLK_s:
	  pattern = solid;
	  break;
	case SDLK_v:
	  pattern = vert_lines;
	  break;
	case SDLK_h:
	  pattern = horiz_lines;
	  break;
	case SDLK_c:
	  pattern = crosshatch;
	  break;
	case SDLK_UP:
	  v_offset = v_offset - 1;
	  if (v_offset < 0)
	    v_offset = pitch - 1;
	  break;
	case SDLK_DOWN:
	  v_offset = v_offset + 1;
	  if (v_offset >= pitch)
	    v_offset = 0;
	  break;
	case SDLK_LEFT:
	  h_offset = h_offset - 1;
	  if (h_offset < 0)
	    h_offset = pitch - 1;
	  break;
	case SDLK_RIGHT:
	  h_offset = h_offset + 1;
	  if (h_offset >= pitch)
	    h_offset = 0;
	  break;
	case SDLK_MINUS:
	case SDLK_KP_MINUS:
	  if ((pattern != solid) && (pitch > 2))
	    pitch--;
	  break;
	case SDLK_PLUS:
	case SDLK_EQUALS:
	case SDLK_KP_PLUS:
	  set_pitch (pitch+1);
	  break;
	case SDLK_0: case SDLK_KP0: set_pitch (  4); break;
	case SDLK_1: case SDLK_KP1: set_pitch (  8); break;
	case SDLK_2: case SDLK_KP2: set_pitch ( 16); break;
	case SDLK_3: case SDLK_KP3: set_pitch ( 32); break;
	case SDLK_4: case SDLK_KP4: set_pitch ( 64); break;
	case SDLK_5: case SDLK_KP5: set_pitch ( 96); break;
	case SDLK_6: case SDLK_KP6: set_pitch (128); break;
	case SDLK_7: case SDLK_KP7: set_pitch (192); break;
	case SDLK_8: case SDLK_KP8: set_pitch (256); break;
	case SDLK_9: case SDLK_KP9: set_pitch (320); break;
	default:
	  break;
	}
    }
}


int enumerate_sdl_modes (int *width, int *height, FILE *out)
{
  SDL_Rect **modes;
  int w = -1;
  int h = -1;
  int m = -1;
  int i;

  modes = SDL_ListModes (NULL, SDL_HWSURFACE | SDL_FULLSCREEN);

  if (! modes)
    fatal (2, "no SDL display modes available\n");

  if (modes == (SDL_Rect **) -1)
    fatal (2, "all SDL resolutions available (on hardware?)\n");

  for (i = 0; modes [i]; i++)
    {
      if (((*width) >= 0) & (modes [i]->w != *width))
	continue;
      if (((*height) >= 0) & (modes [i]->h != *height))
	continue;
      if (out)
	fprintf (out, "  mode %d: %d x %d\n", i, modes [i]->w, modes [i]->h);
      if ((modes [i]->w >= w) && (modes [i]->h >= h))
	{
	  w = modes [i]->w;
	  h = modes [i]->h;
	  m = i;
	}
    }

  if (m < 0)
    fatal (2, "no matching SDL video modes found\n");

  *width = w;
  *height = h;
  return (m);
}


int main (int argc, char *argv [])
{
  SDL_Surface *screen;
  bool fullscreen = true;
  uint32_t sdl_flags;
  int width = -1;
  int height = -1;

  progname = argv [0];

  if (SDL_Init (SDL_INIT_VIDEO) < 0)
    fatal (2, "SDL initialization error %s\n", SDL_GetError ());
  atexit (SDL_Quit);

  while (--argc)
    {
      argv++;
      if ((strcmp (argv [0], "--help") == 0) ||
	  (strcmp (argv [0], "-?") == 0))
	{
	  usage (stderr);
	  help (stderr);
	  exit (0);
	}
      else if (strcmp (argv [0], "--list-modes") == 0)
	{
	  fprintf (stderr, "SDL video modes:\n");
	  enumerate_sdl_modes (& width, & height, stderr);
	  exit (0);
	}
      else if ((strcmp (argv [0], "--window") == 0) ||
	       (strcmp (argv [0], "-w") == 0))
	{
	  fullscreen = false;
	  if (width < 0)
	    width = 320;
	  if (height < 0)
	    height = 240;
	}
      else if (strcmp (argv [0], "-x") == 0)
	{
	  if (--argc == 0)
	    fatal (1, "-x option must be followed by resolution\n");
	  width = atoi (argv [1]);
	  argv++;
	}
      else if (strcmp (argv [0], "-y") == 0)
	{
	  if (--argc == 0)
	    fatal (1, "-y option must be followed by resolution\n");
	  height = atoi (argv [1]);
	  argv++;
	}
      else
	fatal (1, "unrecognized option '%s'\n", argv [0]);
    }

  if (fullscreen)
    {
      enumerate_sdl_modes (& width, & height, NULL);
      printf ("using video resolution %d x %d\n", width, height);
      sdl_flags = SDL_HWSURFACE | SDL_FULLSCREEN;
    }
  else
    sdl_flags = SDL_SWSURFACE;

  screen = SDL_SetVideoMode (width, height, 32, sdl_flags);
  if (! screen)
    fatal (2, "can't set video mode: %s\n", SDL_GetError ());

  if (fullscreen)
    SDL_ShowCursor (SDL_DISABLE);

  main_loop (screen);

  exit (0);
}
